%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%     Perturbed consensus          %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
clc; close all; clear all;
warning off all
%
% figures
font_size = 24;
font_name = 'Times New Roman';
%
% parameters
simulation = 1;
%
% time of simulation
Tmax = 1000;
stepSize = 0.01;
%
% number of agents
N = 8;
%
% Delta operator: parameters
alpha_min = .5;
alpha = alpha_min + (1 - alpha_min)*rand(N,1);
% subsystems
%
% initial conditions
x_0 = [5*randn(N,1)];
%
% interconnection matrix
B = eye(N) - CirculantMatrix(N,1);
[V, J] = eig(B'*B);
%
% storage allocation
x = zeros(N,Tmax);
x(:,1) = x_0;
y = zeros(N, Tmax);
u_shouldbe = zeros(N, Tmax);
u_applied = zeros(N, Tmax);
% Gamma operator: parameters
fCoeff = 1/4; 
fEta = 1 - fCoeff;
fBeta = 1+fCoeff;
%
for t = 1:Tmax
    % simulation
    y(:,t) = x(:,t);
    % 
    currentPhi = zeros(N,1);
    currentDifferences = B*y(:,t);
    for iEdge = 1:N
        currentPhi(iEdge) = currentDifferences(iEdge) + sin(currentDifferences(iEdge)*fCoeff);
    end
    u_shouldbe(:,t) = -B'*currentPhi;
    for agent = 1:N
        u_applied(agent,t) = u_shouldbe(agent, t) + DeltaQuasiSaturationOperator(u_shouldbe(agent,t), .1, alpha(agent));
    end
    x(:, t+1) = x(:,t) + stepSize*(u_applied(:,t));
end
%% figure
times = stepSize*[1:Tmax];
%
figure
plot(times, y, 'k', 'LineWidth', 2)
grid on
xlabel('Time', 'Interpreter', 'Latex', 'FontName', font_name, 'FontSize', font_size)
title('Outputs of the subsystems', 'FontSize', font_size, 'FontName', font_name)
set(gca, 'FontSize', font_size, 'FontName', font_name)
xlim([times(1) times(end)])
%
figure
h_fig = gcf;
plot(times, u_applied(1,:), 'k', 'LineWidth', 2)
hold on
plot(times, u_shouldbe(1,:), '--k')
plot(times, u_shouldbe(2:8,:), '--k')
plot(times, u_applied(2:8,:), 'k', 'LineWidth', 2)
hold on
plot(times, u_shouldbe,'--k')
xlim([times(1) times(end)])
legend('Ideal', 'Perturbed', 'Location', 'NorthWest')
grid on
xlabel('Time', 'Interpreter', 'Latex', 'FontName', font_name, 'FontSize', font_size)
set(gca, 'FontSize', font_size, 'FontName', font_name)
set(gca, 'FontSize', font_size, 'FontName', font_name)
title('Control of the subsystems', 'FontSize', font_size, 'FontName', font_name)
%
%% check of conditions
afNonZeroEigenvaluesOfL = diag(J);
afNonZeroEigenvaluesOfL = afNonZeroEigenvaluesOfL(find(abs(afNonZeroEigenvaluesOfL) > 10^-5));
%
afNormalizedFrequencies = 5*[0:516].^4/516^4;
afFrequencyResponseNominalSystem = freqs(1,[1 0], afNormalizedFrequencies);
afFrequencyResponseNominalSystem(1) = 100000;
%
MultiplierGamma = zeros(2,2,numel(afNormalizedFrequencies));
MultiplierDelta = zeros(2,2,numel(afNormalizedFrequencies));
%
alpha_min = alpha_min;
sigma = .1;
% choice of M
a1 = 10; x1 = 1/(a1);
afM = x1./(sqrt(-1)*afNormalizedFrequencies+a1);
%
for iIndexFreq = 1:numel(afNormalizedFrequencies)
    %
    currentFreq = afNormalizedFrequencies(iIndexFreq);
    MultiplierGamma(:,:,iIndexFreq) = ...
        [2/((fBeta-fEta)*max(afNonZeroEigenvaluesOfL))*(1+real(afM(iIndexFreq))),    1+afM(iIndexFreq);   ...
        1+afM(iIndexFreq)',           0];
    %
    MultiplierDelta(:,:,iIndexFreq) = ...
        [   2,                                          alpha_min*(1-sqrt(-1)*sigma*currentFreq);   ...
            alpha_min*(1+sqrt(-1)*sigma*currentFreq),   0];
    %
end;%
%
aafMatrixMultiplierGamma = zeros(2,2,numel(afNonZeroEigenvaluesOfL), numel(afNormalizedFrequencies));
aafMatrixMultiplierDelta = zeros(2,2,numel(afNonZeroEigenvaluesOfL), numel(afNormalizedFrequencies));
for iIndexEigs = 1:numel(afNonZeroEigenvaluesOfL)
    %
    currentEig = afNonZeroEigenvaluesOfL(iIndexEigs);
    afFractionNumH = afFrequencyResponseNominalSystem./(1+fEta*currentEig*afFrequencyResponseNominalSystem);   
    afFractionNumNothing = 1./(1+fEta*currentEig*afFrequencyResponseNominalSystem);
    afFractionNumMinusEtaLambdaH = -fEta*currentEig*afFrequencyResponseNominalSystem./(1+fEta*currentEig*afFrequencyResponseNominalSystem);
    for iIndexFreq = 1:numel(afNormalizedFrequencies)
        aafMatrixMultiplierGamma(:, :, iIndexEigs, iIndexFreq) = [1, 0;afFractionNumH(iIndexFreq), afFractionNumH(iIndexFreq)];
        aafMatrixMultiplierDelta(:, :, iIndexEigs, iIndexFreq) = [0, 1;afFractionNumNothing(iIndexFreq), afFractionNumMinusEtaLambdaH(iIndexFreq)];
    end
    %
end;%
%
aaafMustBePositiveDefinite = zeros(2,2,numel(afNonZeroEigenvaluesOfL), numel(afNormalizedFrequencies));
aafSmallestEigenvalue = zeros(numel(afNonZeroEigenvaluesOfL), numel(afNormalizedFrequencies));
%
for iIndexEigs = 1:numel(afNonZeroEigenvaluesOfL)
    %
    for iIndexFreq = 1:numel(afNormalizedFrequencies)
        %
        aaafMustBePositiveDefinite(:,:,iIndexEigs, iIndexFreq) = ...
        aafMatrixMultiplierGamma(:,:,iIndexEigs, iIndexFreq)'*MultiplierGamma(:,:,iIndexFreq)*aafMatrixMultiplierGamma(:,:,iIndexEigs, iIndexFreq)+...
        aafMatrixMultiplierDelta(:,:,iIndexEigs,iIndexFreq)'*MultiplierDelta(:,:,iIndexFreq)*aafMatrixMultiplierDelta(:,:,iIndexEigs,iIndexFreq);
        %
        %fCheckShouldBeSmall1 = norm(MultiplierGamma(:,:,iIndexFreq)-MultiplierGamma(:,:,iIndexFreq)')
        %fCheckShouldBeSmall2 = norm(MultiplierGamma(:,:,iIndexFreq)-MultiplierGamma(:,:,iIndexFreq)')
        %fCheckShouldBeSmall3 = norm(aaafMustBePositiveDefinite(:,:,iIndexEigs, iIndexFreq)-aaafMustBePositiveDefinite(:,:,iIndexEigs, iIndexFreq)')
        %
        currentEigenvalues = eig(aaafMustBePositiveDefinite(:,:,iIndexEigs, iIndexFreq));
        aafSmallestEigenvalue(iIndexEigs, iIndexFreq) = min(real(currentEigenvalues));
        %
    end;%
end
%
disp('Smallest eigenvalue of the LHS, condition (iii), Corollary 2, for each nonzero eigenvalue of the Laplacian and for each frequency')
disp('Synchronization of the system is proved if these values are all positive')

SmallestEigenvalues = aafSmallestEigenvalue